# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def math_postprocessor(input_string: str) -> str:
    """
    Parameters
    ----------
    
    input_string: str
        The string from which the numerical answer will be extracted. Only the integer part is extracted.

    Returns
    -------
    str
        The postprocessed string containing the integer part of the answer.
    """
    result = ""
    for char in input_string:
        if char.isdigit():
            result += char
        elif char == ".":
            break
    return result



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import optuna
from typing import Any, Dict, List, Tuple

from sklearn.metrics import (
    fbeta_score,
    balanced_accuracy_score,
    accuracy_score,
    roc_auc_score,
    log_loss,
)

optuna.logging.set_verbosity(optuna.logging.WARNING)


class Tuner:
    def __init__(self) -> None:
        """
        Class for tuning weights and threshold for BSDetector and UQEnsemble
        """
        self.objective_to_func = {
            "fbeta_score": self._f_score,
            "accuracy_score": accuracy_score,
            "balanced_accuracy_score": balanced_accuracy_score,
            "log_loss": log_loss,
            "roc_auc": roc_auc_score,
        }

    def tune_threshold(
        self,
        y_scores: List[float],
        correct_indicators: List[bool],
        thresh_objective: str = "fbeta_score",
        fscore_beta: float = 1,
        bounds: Tuple[float, float] = (0, 1),
        step_size: int = 0.01,
    ) -> float:
        """Conducts 1-dimensional grid search for threshold

        Parameters
        ----------
        y_scores : list of floats
            List of confidence scores

        correct_indicators : list of bool
            A list of boolean indicators of whether self.original_responses are correct.

        thresh_objective: {'fbeta_score', 'accuracy_score', 'balanced_accuracy_score', 'roc_auc', 'log_loss'}, default='fbeta_score'
            Objective function for threshold optimization via grid search.

        fscore_beta : float, default=1
            Value of beta in fbeta_score

        bounds : tuple of floats, default=(0,1)
            Bounds to search for threshold

        step_size : float, default=0.01
            Indicates step size in grid search, if used

        Returns
        -------
        float
            Optimized threshold
        """

        self.fscore_beta = fscore_beta
        threshold_tuning_objective = self.objective_to_func[thresh_objective]
        threshold_values = np.arange(bounds[0], bounds[1], step=step_size)

        y_scores_array = np.array(y_scores)
        y_pred_matrix = (y_scores_array[:, np.newaxis] > threshold_values).astype(int)
        values = -np.array(
            [
                threshold_tuning_objective(correct_indicators, y_pred)
                for y_pred in y_pred_matrix.T
            ]
        )

        best_index = np.argmin(values)
        best_threshold = threshold_values[best_index]
        return best_threshold

    def tune_params(
        self,
        score_lists: List[List[float]],
        correct_indicators: List[bool],
        weights_objective: str = "roc_auc",
        thresh_objective: str = "fbeta_score",
        thresh_bounds: Tuple[float, float] = (0, 1),
        n_trials: int = 100,
        step_size: float = 0.01,
        fscore_beta: float = 1,
    ) -> Dict[str, Any]:
        """
        Tunes weights and threshold parameters on a set of user-provided graded responses.

        Parameters
        ----------
        score_lists : list of lists of floats
            A list of lists of floats. Each interior list is a list of component-specific scores

        correct_indicators : list of bool
            A list of boolean indicators of whether self.original_responses are correct.

        weights_objective : {'fbeta_score', 'accuracy_score', 'balanced_accuracy_score', 'roc_auc', 'log_loss'}, default='roc_auc'
            Objective function for optimization of weights. Must match thresh_objective if one of 'fbeta_score',
            'accuracy_score', 'balanced_accuracy_score'. If same as thresh_objective, joint optimization will be done.
            
        thresh_objective : {'fbeta_score', 'accuracy_score', 'balanced_accuracy_score', 'roc_auc', 'log_loss'}, default='fbeta_score'
            Objective function for threshold optimization via grid search.

        thresh_bounds : tuple of floats, default=(0,1)
            Bounds to search for threshold

        n_trials : int, default=100
            Indicates how many candidates to search over with optuna optimizer

        step_size : float, default=0.01
            Indicates step size in grid search, if used

        fscore_beta : float, default=1
            Value of beta in fbeta_score

        Returns
        -------
        Dict
            Dictionary containing optimized weights and threshold
        """
        self.score_lists = score_lists
        self.k = len(score_lists)
        self.correct_indicators = correct_indicators
        self.weights_objective = weights_objective
        self.thresh_bounds = thresh_bounds
        self.thresh_objective = thresh_objective
        self.n_trials = n_trials
        self.step_size = step_size
        self.fscore_beta = fscore_beta
        self.optimize_jointly = weights_objective == thresh_objective
        self.obj_multiplier = 1 if weights_objective=="logloss" else -1

        # Validate inputs are correct
        self._validate_tuning_inputs()
        self.weights_tuning_objective = self.objective_to_func[self.weights_objective]
        self.threshold_tuning_objective = self.objective_to_func[self.thresh_objective]

        params = self._optimize_objective()
        return {"weights": params[:-1], "thresh": params[-1]}

    def _optimize_objective(self):
        """Runs optimization routine as specified by user"""
        if self.optimize_jointly:
            if self.k > 2:
                print("Jointly optimizing weights and threshold...")
                study = optuna.create_study()
                study.optimize(self._optuna_objective, n_trials=self.n_trials)
                params = tuple(study.best_params.values())
                best_weights = self._normalize_weights(params[:-1])
                return tuple(best_weights) + (params[-1],)
            else:
                print("Jointly optimizing weights and threshold with grid search...")
                params = self._grid_search_weights_thresh()
                best_weights = self._normalize_weights(params[:-1])
                return tuple(best_weights) + (params[-1],)
        else:
            if self.k > 3:
                print("Optimizing weights...")
                study = optuna.create_study()
                study.optimize(self._optuna_objective, n_trials=self.n_trials)
                best_weights_raw = tuple(study.best_params.values())
                best_weights = self._normalize_weights(best_weights_raw)
            else:
                print("Optimizing weights with grid search...")
                best_weights = self._grid_search_weights()
            
            print("Optimizing threshold with grid search...")
            new_scores = self._update_scores(best_weights)
            best_threshold = self.tune_threshold(
                y_scores=new_scores,
                correct_indicators=self.correct_indicators,
                thresh_objective=self.thresh_objective,
                fscore_beta=self.fscore_beta,
            )
            return tuple(best_weights) + (best_threshold,)

    def _f_score(self, y_true, y_pred):
        """Helper function to compute f-beta score"""
        return fbeta_score(y_true, y_pred, beta=self.fscore_beta)

    def _validate_tuning_inputs(self):
        """Helper function to validate tuning inputs"""
        if self.k == 1:
            raise ValueError(
                """Tuning only applies if more than scorer component is present."""
            )
        
        if self.weights_objective not in self.objective_to_func:
            raise ValueError(
                """
            Only 'fbeta_score', 'accuracy_score', 'balanced_accuracy_score', 'roc_auc_score', and 'log_loss' are supported for tuning objectives.
            """
            )
        if self.thresh_objective not in [
            "fbeta_score",
            "accuracy_score",
            "balanced_accuracy_score",
        ]:
            raise ValueError(
                """
                Only 'fbeta_score', 'accuracy_score', 'balanced_accuracy_score' are supported for tuning objectives.
                """
            )
        if self.weights_objective in [
            "fbeta_score",
            "accuracy_score",
            "balanced_accuracy_score",
        ]:
            if not self.optimize_jointly:
                raise ValueError(
                    """
                    thresh_objective must match weights_objective for any threshold-dependent weights_objective.
                    """
                )

    def _optuna_objective(self, trial) -> float:
        """Helper function to define optuna objective"""
        thresh = None
        raw_weights = [
            trial.suggest_float(
                f"weight_{i}", 0, 1
            )
            for i in range(self.k)
        ]
        weights = self._normalize_weights(raw_weights)
        if self.optimize_jointly:
            thresh = trial.suggest_float(
                "thresh", self.thresh_bounds[0], self.thresh_bounds[1]
            )
        ensemble_scores = self._compute_ensemble_scores(weights=weights, score_lists=self.score_lists)
        return self._evaluate_objective(
            y_true=self.correct_indicators, y_pred=ensemble_scores, thresh=thresh
        )

    def _evaluate_objective(self, y_true, y_pred, thresh=None):
        """Helper function to define evaluate objective function for weights"""
        if thresh is not None:
            y_pred = [(y_pred_i > thresh) * 1 for y_pred_i in y_pred]
        return self.obj_multiplier * self.weights_tuning_objective(y_true, y_pred)

    def _update_scores(self, weights: List[float]) -> List[float]:
        """Update confidence scores"""
        return self._compute_ensemble_scores(weights, score_lists=self.score_lists)

    def _compute_ensemble_scores(
        self, weights: List[float], score_lists: List[List[float]]
    ) -> List[float]:
        """Helper function to compute dot product for getting ensemble scores"""
        final_scores = []
        for scores_i in zip(*score_lists):
            filtered_scores_weights = [
                (si, w) for si, w in zip(scores_i, weights) if not np.isnan(si)
            ]
            filtered_scores, filtered_weights = zip(*filtered_scores_weights)
            normalized_weights = self._normalize_weights(filtered_weights)

            weighted_sum = sum(
                si * nw for si, nw in zip(filtered_scores, normalized_weights)
            )
            final_scores.append(weighted_sum)
        return final_scores
    
    def _grid_search_weights_thresh(self):
        """
        Joint grid search optimization for k weights and threshold.
        Use only if k==2, only one weight is free (the second is 1 - w1)    
        """        
        weight_grid = np.linspace(0, 1, int(1 / self.step_size))
        threshold_grid = np.linspace(
            self.thresh_bounds[0] + self.step_size, 
            self.thresh_bounds[1] - self.step_size, 
            int((self.thresh_bounds[1] - self.thresh_bounds[0]) / (self.step_size) - 1)
        )
        best_cost = np.inf
        for w in weight_grid:
            weights = np.array([w, 1-w])  # automatically sum to 1
            for thresh in threshold_grid:
                cost = self._evaluate_objective(
                    y_true=self.correct_indicators,
                    y_pred=self._update_scores(weights),
                    thresh=thresh,
                )
                if cost < best_cost:
                    best_cost = cost
                    best_weights = weights
                    best_thresh = thresh
        return tuple(best_weights) + (best_thresh, )  

    def _grid_search_weights(self):
        """
        Grid search for weights only.
        For k==2: one free weight.
        For k==3: two free weights where the third is 1 - (w1+w2)
        Only consider feasible regions (weights are non-negative and sum to 1).   
        """
        best_cost = np.inf
        if self.k == 2:
            w_grid = np.linspace(0, 1, int(1 / self.step_size))
            for w in w_grid:
                weights = np.array([w, 1-w])
                cost = self._evaluate_objective(
                    y_true=self.correct_indicators,
                    y_pred=self._update_scores(weights),
                )
                if cost < best_cost:
                    best_cost = cost
                    best_weights = weights
        elif self.k == 3:
            w1_grid = np.linspace(0, 1, int(1 / self.step_size))
            w2_grid = np.linspace(0, 1, int(1 / self.step_size))
            for w1 in w1_grid:
                for w2 in w2_grid:
                    w3 = 1 - w1 - w2
                    if w3 < 0:  # infeasible, skip
                        continue
                    weights = np.array([w1, w2, w3])
                    cost = self._evaluate_objective(
                        y_true=self.correct_indicators,
                        y_pred=self._update_scores(weights),
                    )
                    if cost < best_cost:
                        best_cost = cost
                        best_weights = weights
        return tuple(best_weights)

    @staticmethod
    def _normalize_weights(weights: List[float]) -> List[float]:
        """Helper function to ensure weights sum to 1"""
        return [w / np.sum(weights) for w in weights]



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional
import math
import numpy as np
from langchain_core.language_models.chat_models import BaseChatModel

from uqlm.scorers.baseclass.uncertainty import UncertaintyQuantifier, UQResult


class WhiteBoxUQ(UncertaintyQuantifier):
    def __init__(
        self,
        llm: Optional[BaseChatModel] = None,
        system_prompt: str = "You are a helpful assistant.",
        max_calls_per_min: Optional[int] = None,
        scorers: Optional[List[str]] = None,
    ) -> None:
        """
        Class for computing white-box UQ confidence scores. This class offers two confidence scores, normalized 
        probability :footcite:`malinin2021uncertaintyestimationautoregressivestructured` and minimum probability :footcite:`manakul2023selfcheckgptzeroresourceblackboxhallucination`.
        
        Parameters
        ----------
        llm : BaseChatModel
            A langchain llm object to get passed to chain constructor. User is responsible for specifying
            temperature and other relevant parameters to the constructor of their `llm` object.

        max_calls_per_min : int, default=None
            Used to control rate limiting.

        system_prompt : str or None, default="You are a helpful assistant."
            Optional argument for user to provide custom system prompt

        scorers : subset of {
            "imperplexity", "geometric_mean_probability", "min_probability", "max_probability",
        }, default=None
            Specifies which black box (consistency) scorers to include. If None, defaults to all.
        """
        super().__init__(
            llm=llm,
            max_calls_per_min=max_calls_per_min,
            system_prompt=system_prompt,
        )
        self.scorers = scorers if scorers else self.white_box_names

    async def generate_and_score(
        self,
        prompts: List[str],
    ) -> UQResult:
        """
        Generate responses and compute white-box confidence scores based on extracted token probabilities.

        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        Returns
        -------
        UQResult
            UQResult containing prompts, responses, logprobs, and white-box UQ scores
        """
        assert hasattr(self.llm, "logprobs"), """
        BaseChatModel must have logprobs attribute and have logprobs=True
        """
        self.llm.logprobs = True
        responses = await self.generate_original_responses(prompts)
        return self.score(
            prompts=prompts, responses=responses, logprobs_results=self.logprobs,
        )
        
    def score(
        self,
        logprobs_results: List[List[Dict[str, Any]]],
        prompts: Optional[List[str]] = None,
        responses: Optional[List[str]] = None,
    ) -> UQResult:
        """
        Compute white-box confidence scores from provided logprobs.

        Parameters
        ----------
        logprobs_results : list of logprobs_result
            List of dictionaries, each returned by BaseChatModel.agenerate
        
        prompts : list of str, default=None
            A list of input prompts for the model.
            
        responses : list of str, default=None
            A list of model responses for the prompts. 

        Returns
        -------
        UQResult
            UQResult containing prompts, responses, logprobs, and white-box UQ scores
        """        
        
        self.logprobs = logprobs_results
        self.prompts = prompts
        self.responses = responses
        
        data = {}
        if self.prompts:
            data["prompts"] = self.prompts
        if self.responses:
            data["responses"] = self.responses
            
        data['logprobs'] = self.logprobs
        scores = self._compute_scores(self.logprobs)
        for key in self.scorers:
            data[key] = scores[key]

        result = {
            "data": data,
            "metadata": {
                "temperature": None if not self.llm else self.llm.temperature,
            },
        }
        return UQResult(result)
    
    def _compute_scores(self, logprobs_results: List[List[Dict[str, Any]]]) -> List[float]:
        """
        This method computes token-probability-based confidence scores.
        """
        return {
            "normalized_probability": [np.nan if not r else self._norm_prob(r) for r in logprobs_results],
            "min_probability": [np.nan if not r else self._min_prob(r) for r in logprobs_results],
        }

    def _norm_prob(self, logprobs: List[Dict[str, Any]]) -> float:
        """Compute normalized token probability"""
        return math.exp(self.avg_logprob(logprobs))

    def _min_prob(self, logprobs: List[Dict[str, Any]]) -> float:
        """Compute minimum token probability"""
        return min(self._get_probs(logprobs))

    def avg_logprob(self, logprobs: List[Dict[str, Any]]) -> float:
        "Compute average logprob"
        return np.mean(self.get_logprobs(logprobs))
    
    @staticmethod
    def _get_probs(logprobs):
        """Extract token probabilities"""
        return [math.exp(d["logprob"]) for d in logprobs]
    
    @staticmethod
    def get_logprobs(logprobs):
        """Extract log token probabilities"""
        return [d["logprob"] for d in logprobs]


# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from langchain_core.language_models.chat_models import BaseChatModel
from typing import Any, List, Optional

from uqlm.scorers.baseclass.uncertainty import UncertaintyQuantifier, UQResult
from uqlm.black_box import BertScorer, CosineScorer, MatchScorer


class BlackBoxUQ(UncertaintyQuantifier):
    def __init__(
        self,
        llm: Optional[BaseChatModel] = None,
        scorers: Optional[List[str]] = None,
        device: Any = None,
        use_best: bool = True,
        nli_model_name: str = "microsoft/deberta-large-mnli",
        postprocessor: Any = None,
        system_prompt: str = "You are a helpful assistant.",
        max_calls_per_min: Optional[int] = None,
        sampling_temperature: float = 1.0,
        use_n_param: bool = False,
        max_length: int = 2000,
        verbose: bool = False,
    ) -> None:
        """
        Class for black box uncertainty quantification. Leverages multiple responses to the same prompt to evaluate
        consistency as an indicator of hallucination likelihood. 

        Parameters
        ----------
        llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `llm` object.

        scorers : subset of {
            'semantic_negentropy', 'noncontradiction', 'exact_match', 'bert_score', 'bleurt', 'cosine_sim'
        }, default=None
            Specifies which black box (consistency) scorers to include. If None, defaults to
            ["semantic_negentropy", "noncontradiction", "exact_match", "cosine_sim"]. 
            
        device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that NLI model use for prediction. Only applies to 'semantic_negentropy', 'noncontradiction' 
            scorers. Pass a torch.device to leverage GPU.

        use_best : bool, default=True
            Specifies whether to swap the original response for the uncertainty-minimized response
            based on semantic entropy clusters. Only used if `scorers` includes 'semantic_negentropy' or 'noncontradiction'.
            
        nli_model_name : str, default="microsoft/deberta-large-mnli"
            Specifies which NLI model to use. Must be acceptable input to AutoTokenizer.from_pretrained() and 
            AutoModelForSequenceClassification.from_pretrained()

        postprocessor : callable, default=None
            A user-defined function that takes a string input and returns a string. Used for postprocessing
            outputs.

        system_prompt : str or None, default="You are a helpful assistant."
            Optional argument for user to provide custom system prompt

        max_calls_per_min : int, default=None
            Specifies how many api calls to make per minute to avoid a rate limit error. By default, no
            limit is specified.

        sampling_temperature : float, default=1.0
            The 'temperature' parameter for llm model to generate sampled LLM responses. Must be greater than 0.

        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when num_responses > 1.
            
        max_length : int, default=2000
            Specifies the maximum allowed string length. Responses longer than this value will be truncated to 
            avoid OutOfMemoryError
            
        verbose : bool, default=False
            Specifies whether to print the index of response currently being scored.
        """
        super().__init__(
            llm=llm,
            device=device,
            system_prompt=system_prompt,
            max_calls_per_min=max_calls_per_min,
            use_n_param=use_n_param,
            postprocessor=postprocessor,
        )
        self.prompts = None
        self.max_length = max_length
        self.verbose = verbose
        self.use_best = use_best
        self.sampling_temperature = sampling_temperature
        self.nli_model_name = nli_model_name
        self._validate_scorers(scorers)
        self.use_nli = ("semantic_negentropy" in self.scorers) or (
            "noncontradiction" in self.scorers
        )
        if self.use_nli:
            self._setup_nli(nli_model_name)

    async def generate_and_score(
        self,
        prompts: List[str],
        num_responses: int = 5,
    ) -> UQResult:
        """
        Generate LLM responses, sampled LLM (candidate) responses, and compute confidence scores with specified scorers for the provided prompts.

        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        num_responses : int, default=5
            The number of sampled responses used to compute consistency.

        Returns
        -------
        UQResult
            UQResult containing data (prompts, responses, and scores) and metadata
        """
        self.prompts = prompts
        self.num_responses = num_responses

        responses = await self.generate_original_responses(prompts)
        sampled_responses = await self.generate_candidate_responses(prompts)
        return self.score(
            responses = responses, sampled_responses = sampled_responses,
        )

    def score(
        self, responses: List[str], sampled_responses: List[List[str]]
             
    ) -> UQResult:
        """
        Compute confidence scores with specified scorers on provided LLM responses. Should only be used if responses and sampled responses
        are already generated. Otherwise, use `generate_and_score`.

        Parameters
        ----------
        responses : list of str, default=None
            A list of model responses for the prompts. 

        sampled_responses : list of list of str, default=None
            A list of lists of sampled LLM responses for each prompt. These will be used to compute consistency scores by comparing to 
            the corresponding response from `responses`.

        Returns
        -------
        UQResult
            UQResult containing data (prompts, responses, and scores) and metadata
        """
        print("Computing confidence scores...")
        self.responses = responses
        self.sampled_responses = sampled_responses
        self.num_responses = len(sampled_responses[0])
        
        self.scores_dict = {k: [] for k in self.scorer_objects}
        if self.use_nli:
            compute_entropy = ("semantic_negentropy" in self.scorers)
            nli_scores = self.nli_scorer.evaluate(
                responses=self.responses,
                sampled_responses=self.sampled_responses,
                use_best=self.use_best,
                compute_entropy=compute_entropy,
            )
            if self.use_best:
                self.original_responses = self.responses.copy()
                self.responses = nli_scores["responses"]
                self.sampled_responses = nli_scores["sampled_responses"]

            for key in ["semantic_negentropy", "noncontradiction"]:
                if key in self.scorers:
                    if key == "semantic_negentropy":
                        nli_scores[key] = [
                            1 - s
                            for s in self.nli_scorer._normalize_entropy(
                                nli_scores[key]
                            )
                        ]  # Convert to confidence score
                    self.scores_dict[key] = nli_scores[key]

        # similarity scorers that follow the same pattern
        for scorer_key in ["exact_match", "bert_score", "bleurt", "cosine_sim"]:
            if scorer_key in self.scorer_objects:
                self.scores_dict[scorer_key] = self.scorer_objects[scorer_key].evaluate(
                    responses=self.responses,
                    sampled_responses=self.sampled_responses,
                )
                
        return self._construct_result()
                
    def _construct_result(self) -> Any:
        """Constructs UQResult object"""
        data = {
            "responses": self.responses,
            "sampled_responses": self.sampled_responses,
        }
        if self.prompts:
            data["prompts"] = self.prompts
        data.update(self.scores_dict)
        result = {
            "data": data,
            "metadata": {
                "temperature": None if not self.llm else self.llm.temperature,
                "sampling_temperature": None
                if not self.sampling_temperature
                else self.sampling_temperature,
                "num_responses": self.num_responses,
                "scorers": self.scorers,
            },
        }
        return UQResult(result)

    def _validate_scorers(self, scorers: List[Any]) -> None:
        "Validate scorers and construct applicable scorer attributes"
        self.scorer_objects = {}
        if scorers is None:
            scorers = self.default_black_box_names
        for scorer in scorers:
            if scorer == "exact_match":
                self.scorer_objects["exact_match"] = MatchScorer()
            elif scorer == "bert_score":
                self.scorer_objects["bert_score"] = BertScorer()
            elif scorer == "bleurt":
                from uqlm.black_box import BLEURTScorer
                self.scorer_objects["bleurt"] = BLEURTScorer()
            elif scorer == "cosine_sim":
                self.scorer_objects["cosine_sim"] = CosineScorer()
            elif scorer in ["semantic_negentropy", "noncontradiction"]:
                continue
            else:
                raise ValueError(
                    """
                    scorers must be one of ['semantic_negentropy', 'noncontradiction', 'exact_match', 'bert_score', 'bleurt', 'cosine_sim']
                    """
                )
        self.scorers = scorers



# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from typing import List, Optional, Union
from langchain_core.language_models.chat_models import BaseChatModel

from uqlm.judges.judge import LLMJudge
from uqlm.scorers.baseclass.uncertainty import UncertaintyQuantifier, UQResult


class LLMPanel(UncertaintyQuantifier):
    def __init__(
        self,
        judges: List[Union[LLMJudge, BaseChatModel]],
        llm: Optional[BaseChatModel] = None,
        system_prompt: str = "You are a helpful assistant.",
        max_calls_per_min: Optional[int] = None,
    ) -> None:
        """
        Class for aggregating multiple instances of LLMJudge using min, max, or majority voting

        Parameters
        ----------
        judges: list of LLMJudge or BaseChatModel
            Judges to use. If BaseChatModel, LLMJudge is instantiated using default parameters.

        llm : BaseChatModel
            A langchain llm object to get passed to chain constructor. User is responsible for specifying
            temperature and other relevant parameters to the constructor of their `llm` object.

        max_calls_per_min : int, default=None
            Used to control rate limiting. Will be used for original llm and any judges constructed
            from instances of BaseChatModel in judges

        system_prompt : str or None, default="You are a helpful assistant."
            Optional argument for user to provide custom system prompt
        """
        super().__init__(
            llm=llm,
            max_calls_per_min=max_calls_per_min,
            system_prompt=system_prompt,
        )
        self.judges = []
        for judge in judges:
            if isinstance(judge, BaseChatModel):
                judge = LLMJudge(llm=judge, max_calls_per_min=max_calls_per_min)
            elif not isinstance(judge, LLMJudge):
                raise ValueError(
                    "judges must be a list containing instances of either LLMJudge or BaseChatModel"
                )
            self.judges.append(judge)

    async def generate_and_score(
        self,
        prompts: List[str],
    ) -> UQResult:
        """
        Generate LLM responses to provided prompts and use panel of judges to score responses for correctness.

        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        Returns
        -------
        UQResult
            UQResult containing prompts, responses, Q/A concatenations, judge responses, and judge scores
        """
        responses = await self.generate_original_responses(prompts)
        return await self.score(prompts=prompts, responses=responses)
    
    async def score(
        self,
        prompts: List[str],
        responses: Optional[List[str]] = None,
    ) -> UQResult:
        """
        Use panel to of judges to score provided responses for correctness. Use if responses are already generated. Otherwise,
        use generate_and_score.

        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        responses: list of str, default = None
            A list of LLM responses for the corresponding to the provided prompts. 

        Returns
        -------
        UQResult
            UQResult containing prompts, responses, Q/A concatenations, judge responses, and judge scores
        """
        self.prompts = prompts
        self.responses = responses
        data = {
            "prompts": prompts,
            "responses": responses,
        }
        
        judge_count = 1
        scores_lists = []
        for judge in self.judges:
            tmp = await judge.judge_responses(prompts=prompts, responses=responses)
            scores_lists.append(tmp["scores"])
            data[f"judge_{judge_count}"] = tmp["scores"]
            judge_count += 1

        scores_dict = {
            "avg": [np.mean(scores) for scores in zip(*scores_lists)],
            "max": [np.max(scores) for scores in zip(*scores_lists)],
            "min": [np.min(scores) for scores in zip(*scores_lists)],
            "median": [np.median(scores) for scores in zip(*scores_lists)],
        }
        data.update(scores_dict)
        result = {
            "data": data,
            "metadata": {
                "num_judges": len(self.judges),
                "temperature": None if not self.llm else self.llm.temperature,
            },
        }
        return UQResult(result)


# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib.resources as resources
import numpy as np
import os
import requests
import zipfile
from requests.exceptions import RequestException
from zipfile import BadZipFile

from typing import List

from uqlm.black_box.baseclass.similarity_scorer import SimilarityScorer


class BLEURTScorer(SimilarityScorer):
    def __init__(self) -> None:
        """
        Class for computing BLEURT Scores between original responses and candidates. For more on
        BLEURT, refer to Sellam et al.(2020) :footcite:`sellam2020bleurtlearningrobustmetrics`. Requires
        installation of `bleurt` package. Install using:
        `pip install pip install --user git+https://github.com/google-research/bleurt.git`

        Raises
        ------
        RuntimeError
            If there's an error downloading or initializing the BLEURT checkpoint
        """
        try:
            from bleurt.score import BleurtScorer
        except ImportError:    
            raise ImportError(
            """
            The bleurt package is required to use BLEURTScorer but is not installed. Please install it using:\n
            `pip install pip install --user git+https://github.com/google-research/bleurt.git`
            """
            )
        try:
            checkpoint = self._set_bleurt_checkpoint()
            self.bleurt_scorer = BleurtScorer(checkpoint)
        except Exception as e:
            raise RuntimeError(
                f"Failed to initialize BLEURT scorer. Error: {str(e)}"
            ) from e

    def evaluate(
        self, responses: List[str], sampled_responses: List[List[str]]
    ) -> List[float]:
        """
        This method computes model-based text similarity metrics values for the provided pairs of texts.

        Parameters
        ----------
        responses : list of strings
            Original LLM response

        sampled_responses : list of list of strings
            Candidate responses to be compared to the original response

        Returns
        -------
        List of float
            Mean BLEURT scores
        """
        return [
            self._compute_score(response=responses[i], candidates=sampled_responses[i])
            for i in range(len(responses))
        ]

    def _compute_score(self, response: str, candidates: List[str]) -> float:
        """Compute BLEURT scores between a response and candidate responses"""
        duplicated_response = [response] * len(candidates)
        return np.mean(
            self.bleurt_scorer.score(
                references=duplicated_response, candidates=candidates
            )
        )

    def _set_bleurt_checkpoint(self):
        """Sets up checkpoint"""        
        resource_path = resources.files("uqlm.resources").joinpath("BLEURT-20")
        if not resource_path.is_dir():
            self._download_and_unzip(
                url="https://storage.googleapis.com/bleurt-oss-21/BLEURT-20.zip",
                my_file_path=resources.files("uqlm.resources"),
            )
        return resource_path

    @staticmethod
    def _download_and_unzip(url, my_file_path):
        """Download BLEURT checkpoint, unzip, and delete zip file"""
        zip_file_path = os.path.join(my_file_path, "BLEURT-20.zip")
        print(f"BLEURT checkpoint not found. Downloading to: {zip_file_path}")

        try:
            response = requests.get(url)
            if response.status_code == 200:
                with open(zip_file_path, "wb") as f:
                    f.write(response.content)
                print(f"Downloaded zip file to: {zip_file_path}")
            else:
                print(f"Failed to download file. Status code: {response.status_code}")
                return
        except RequestException as e:
            print(f"Network error occurred while downloading BLEURT checkpoint: {str(e)}")
            return

        try:
            print(f"Unzipping: {zip_file_path}")
            with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
                zip_ref.extractall(my_file_path)
                print(f"Unzipped files to: {my_file_path}")
        except BadZipFile as e:
            print(f"Error: The downloaded BLEURT zip file is corrupted: {str(e)}")
        except Exception as e:
            print(f"Unexpected error while extracting BLEURT zip file: {str(e)}")
        finally:
            if os.path.exists(zip_file_path):
                os.remove(zip_file_path)
                print(f"Deleted zip file: {zip_file_path}")
            return